//All files to include
#include <stdint.h>
//Constants
const VRefSet = 5000;

//Regular variables
unsigned int ADCIn = 0, Vdd = 0, VRefIn = 0;

// RS232
char txt[16];
void PrintHandler(char c)
{
  UART1_Write(c);
}

void UART1_jump_a_line()
{
 UART1_Write(10);
 UART1_Write(13);
}

//USB Constants
const unsigned int USBRamStart=0x500;
const unsigned short USBInDataLength=64, USBOutDataLength=64;

// Buffer of 64 bytes in USB RAM
char buffer[64] absolute 0x500 ;

//USB Variables
bit USBHasEnteredSuspendMode;
unsigned short USBCount;
unsigned short USBInData[USBInDataLength] absolute USBRamStart+USBInDataLength;
unsigned short USBOutData[USBOutDataLength] absolute USBRamStart;

//Declare needed functions
#define Lo(param) ((char *)&param)[0]
#define Hi(param) ((char *)&param)[1]
#define Higher(param) ((char *)&param)[2]
#define Highest(param) ((char *)&param)[3]
void main();

void ADC_Init() {
     //Turn on ADC Module
     ADCON0.ADON = 1;

     //Set VDD as positive voltage refrence
     ADCON1.VCFG0 = 0;

     //Set VSS as negative voltage refrence
     ADCON1.VCFG1 = 0;

     //Set ADC Conversion clock to Fosc/64 (for 48MHz) so TAD~=1.33us
     ADCON2.ADCS2 = 1;
     ADCON2.ADCS1 = 1;
     ADCON2.ADCS0 = 0;

     //ADC needs about 2.45us so TACQ = 2TAD for 2.66us
     ADCON2.ACQT2 = 0;
     ADCON2.ACQT1 = 0;
     ADCON2.ACQT0 = 1;

     //Right justify ADC results
     ADCON2.ADFM = 1;
}

unsigned int ADC_Get_Sample(unsigned short Channel) 
{
// PIC 18F4550  :  ADC 10 Bits
         unsigned int Result = 0;

         //Select channel and enable ADC
         ADCON0 = (Channel << 2) + 1;

         //Start ADC conversion then wait until it is finished
         ADCON0.GO = 1;
         while (ADCON0.GO);

         //Right justified calculation
         Lo(Result) = ADRESL;
         Hi(Result) = ADRESH;

         return Result;
}

//Prepare to goto sleep
void USBEnterSuspendMode() {
        USBHasEnteredSuspendMode = 1;
}

//Prepare to wake up from sleep
void USBExitSuspendMode() {
        USBHasEnteredSuspendMode = 0;
}

void interrupt() {
     //USB Interrupt
     USB_Interrupt_Proc();
}

void USBDev_CDCParamsChanged(){

}

uint8_t dataReceived = 0;
uint16_t dataReceivedSize;

void interrupt_low() {
}

void main() 
{
     // Orange : RC6 - TX
     // Blanc  : RC7 - RX
     UART1_Init(115200); // Initialize UART module at 115200 bps
     //Initialize variables
     USBHasEnteredSuspendMode = 0;     
     for (USBCount = 0; USBCount < 64; USBCount++) {
         USBInData[USBCount] = USBOutData[USBCount] = 255;
     }

     //Initialize pins
     TRISA = 0;              //Set PORTA as output
     TRISB = 0;              //Set PORTB as output
     TRISC = 0;              //Set PORTC as output
     TRISD = 0;              //Set PORTD as output
     TRISE = 0;              //Set PORTE as output
     TRISB.B4 = 1;           //RB4 switch                
     RBPU_bit = 0;           //Enable weak pull-up on all ports

     //ADC Setup and Enable for RA0
     TRISA.TRISA0 = 1;
     ADCON1 = 0b001110;
     ADC_Init();

     //Enable USB
     HID_Enable(&USBInData, &USBOutData);

     //Assume a VRefIn value since there is no FVR
     VRefIn = 4095;

     //Calculate VDD based on the VRef reading
     Vdd = (unsigned long int) 4095 * VRefSet / VRefIn;

     //Main process loop
     while(1)
     {
           //Delay_ms(25);
           if(UCON.SUSPND) {
                if(!USBHasEnteredSuspendMode) {
                        //Configurations for going to sleep
                        USBEnterSuspendMode();
                } else {
                        //Process if USB is not connected or if PC is in sleep state (S1, S3, S4, S5)
                        PORTC = 85;
                        Delay_ms(25);
                        PORTC = 170;
                        Delay_ms(25);
                        
                        //Don//t continue with the rest of the code
                        continue;
                }
           } else {
                if (USBHasEnteredSuspendMode) {
                   USBExitSuspendMode();
                }
           }     

           //LED indicator to see if the device has been programmed properly
           PORTB.RB5 = 1;
           //Read channel 0 (RA0) and output the result into ADCIn variable
           ADCIn = ADC_Get_Sample(0);
           PORTD = ADCIn >> 2;   // PIC 18F4550
           // http://www.farrellf.com/TelemetryViewer/
           // https://youtu.be/hooo9Ro06Bs
           // ac:telemetry
           //UART1_Write_Text("ADCin: ");
           IntToStr(ADCIn, txt);
           UART1_Write_Text(txt);  // sends back text
           UART1_jump_a_line();
           PORTB.RB5 = 0;
           //Write to USB if data was read
           if(HID_Read())
           {
              USBOutData[0] = 12;        //ADC Resolution
              USBOutData[1] = Hi(ADCIn);
              USBOutData[2] = Lo(ADCIn);
              USBOutData[3] = Hi(Vdd);
              USBOutData[4] = Lo(Vdd);
              USBOutData[5] = Hi(VRefIn);
              USBOutData[6] = Lo(VRefIn);
              HID_Write(&USBOutData, USBOutDataLength);
           }
     }
     //Disable USB
     HID_Disable();
}